// Baixa Tot native logic

#include <iostream>
#include <stdint.h>
#include <string.h>
#include <unistd.h>
#include <json/json.h>
#include <sys/stat.h>

extern "C" {
#include <bestiola/base64.h>
}

union msg_size {
	uint32_t u;
	char c[4];
};

static bool s_debug = true;
static FILE *s_stream;

static void log(const std::string &s) {
	if (s_debug)
		std::cerr << s << std::endl;
}

static void *base64_decode_buffer(const char *s, size_t *dsize) {
	if (!s || !*s)
		return NULL;

	size_t slen = strlen(s);
	if (slen % 4)
		return NULL;

	size_t decoded_size = slen / 4 * 3;
	if (s[slen - 1] == '=') {
		if (s[slen - 2] == '=') {
			slen -= 2;
			decoded_size -= 2;
		}
		else {
			slen--;
			decoded_size--;
		}
	}

	for (size_t i = 0; i < slen; i++) {
		if (!is_base64(s[i]))
			return NULL;
	}

	void *decoded = malloc(decoded_size);
	if (!decoded)
		return NULL;

	if (dsize)
		*dsize = decoded_size;

	const char *e = s;
	char *d = (char *) decoded;
	for (size_t i = 0; i < decoded_size; i += 3) {
		unsigned char v[4];
		switch (decoded_size - i) {
		case 1:
			for (int j = 0; j < 2; j++)
				v[j] = base64_value(*e++);
			*d++ = v[0] << 2 | v[1] >> 4;
			break;
		case 2:
			for (int j = 0; j < 3; j++)
				v[j] = base64_value(*e++);
			*d++ = v[0] << 2 | v[1] >> 4;
			*d++ = v[1] << 4 | v[2] >> 2;
			break;
		default:
			for (int j = 0; j < 4; j++)
				v[j] = base64_value(*e++);
			*d++ = v[0] << 2 | v[1] >> 4;
			*d++ = v[1] << 4 | v[2] >> 2;
			*d++ = v[2] << 6 | v[3];
		}
	}
	return decoded;
}

static bool close_file(Json::Value &req, Json::Value &res) {
	if (fclose(s_stream))
		return true;

	res["result"] = "ok";
	return false;
}

static bool initialize(Json::Value &req, Json::Value &res) {
	s_debug = req["debug"].asBool();

	const char *home = getenv("HOME");
	if (!home) {
		log("Could not get home directory");
		return true;
	}

	char *dir;
	if (asprintf(&dir, "%s/.baixa-tot", home) == -1)
		return true;

	int ret = mkdir(dir, S_IRWXU);
	if (ret && errno != EEXIST) {
		log("Could not create directory");
		free(dir);
		return true;
	}

	ret = chdir(dir);
	free(dir);
	if (ret) {
		log("Could not change directory");
		return true;
	}

	res["result"] = "ack";
	return false;
}

static bool open_file(Json::Value &req, Json::Value &res) {
	const char *url = req["url"].asCString();

	const char *slash = NULL;
	const char *end = strchr(url, '?');
	if (end) {
		for (end--; end > url; end--) {
			if (*end == '/') {
				slash = end;
				break;
			}
		}
	}
	else
		slash = strrchr(url, '/');

	if (!slash)
		return true;

	struct timespec tp;
	if (clock_gettime(CLOCK_REALTIME, &tp))
		return true;

	char *pathname;
	if (asprintf(&pathname, "%ld.%09ld-%d-%s", tp.tv_sec, tp.tv_nsec,
		getpid(), slash + 1) == -1)
		return true;

	long name_max = pathconf(".", _PC_NAME_MAX);
	if (name_max > 0 && strlen(pathname) > (unsigned long) name_max)
		pathname[name_max] = '\0';

	for (char *c = pathname; *c; c++) {
		if (*c == '/')
			*c = '_';
	}

	int ret = access(pathname, F_OK);
	if (!ret || errno != ENOENT) {
		free(pathname);
		return true;
	}

	s_stream = fopen(pathname, "w");
	free(pathname);
	if (!s_stream) {
		log("Could not open file");
		return true;
	}

	res["result"] = "ack";
	return false;
}

static bool write_data(Json::Value &req, Json::Value &res) {
	const char *data = req["data"].asCString();

	size_t dsize;
	void *buf = base64_decode_buffer(data, &dsize);
	if (!buf) {
		log("Could not decode data");
		return true;
	}

	size_t ret = fwrite(buf, 1, dsize, s_stream);
	free(buf);
	if (ret != dsize) {
		log("Could not write data");
		return true;
	}

	res["result"] = "ack";
	return false;
}

static bool read_request(Json::Value &req, bool &end) {
	msg_size size;
	end = fread(size.c, 4, 1, stdin) != 1;
	if (end)
		return ferror(stdin);
	char *buf = new char[size.u + 1];
	bool error = fread(buf, 1, size.u, stdin) != size.u;
	buf[size.u] = '\0';

	if (!error) {
		Json::Reader reader;
		if (!reader.parse(buf, buf + size.u, req))
			error = true;
	}

	delete[] buf;

	return error;
}

static bool write_response(Json::Value &res) {
	Json::FastWriter writer;
	writer.omitEndingLineFeed();
	std::string str = writer.write(res);
	msg_size size;
	size.u = str.length();

	if (fwrite(size.c, 4, 1, stdout) != 1)
		return true;
	if (fwrite(str.c_str(), 1, size.u, stdout) != size.u)
		return true;
	return fflush(stdout);
}

static bool web_extension_protocol(void) {
	for (;;) {
		bool end;
		Json::Value req;
		if (read_request(req, end))
			return true;
		if (end)
			return false;

		bool error = false;
		Json::Value res;
		std::string command = req["command"].asString();
		if (command == "close_file")
			error = close_file(req, res);
		else if (command == "initialize")
			error = initialize(req, res);
		else if (command == "open_file")
			error = open_file(req, res);
		else if (command == "write_data")
			error = write_data(req, res);
		else {
			log("Unknown command " + command);
			error = true;
		}

		if (error) {
			res.clear();
			res["result"] = "error";
		}

		if (write_response(res) || error)
			return true;
	}
}

int main(void) {
	bool error = false;

	if (web_extension_protocol())
		error = true;

	return error? EXIT_FAILURE: EXIT_SUCCESS;
}
